import traceback
import logging
import socket
import sys
from datetime import datetime
try:
    import json
except ImportError:
    import simplejson as json


class LogstashFormatterBase(logging.Formatter):

    def __init__(self, message_type='Logstash', tags=None, fqdn=False):
        self.message_type = message_type
        self.tags = tags if tags is not None else []

        if fqdn:
            self.host = socket.getfqdn()
        else:
            self.host = socket.gethostname()

    def get_extra_fields(self, record):
        # The list contains all the attributes listed in
        # http://docs.python.org/library/logging.html#logrecord-attributes
        skip_list = (
            'args', 'asctime', 'created', 'exc_info', 'exc_text', 'filename',
            'funcName', 'id', 'levelname', 'levelno', 'lineno', 'module',
            'msecs', 'msecs', 'message', 'msg', 'name', 'pathname', 'process',
            'processName', 'relativeCreated', 'thread', 'threadName', 'extra',
            'auth_token', 'password', 'stack_info')

        if sys.version_info < (3, 0):
            easy_types = (basestring, bool, dict, float, int, long, list, type(None))
        else:
            easy_types = (str, bool, dict, float, int, list, type(None))

        fields = {}

        for key, value in record.__dict__.items():
            if key not in skip_list:
                if isinstance(value, easy_types):
                    fields[key] = value
                else:
                    fields[key] = repr(value)

        return fields

    def get_debug_fields(self, record):
        fields = {
            'stack_trace': self.format_exception(record.exc_info),
            'lineno': record.lineno,
            'process': record.process,
            'thread_name': record.threadName,
        }

        # funcName was added in 2.5
        if not getattr(record, 'funcName', None):
            fields['funcName'] = record.funcName

        # processName was added in 2.6
        if not getattr(record, 'processName', None):
            fields['processName'] = record.processName

        return fields

    @classmethod
    def format_source(cls, message_type, host, path):
        return "%s://%s/%s" % (message_type, host, path)

    @classmethod
    def format_timestamp(cls, time):
        tstamp = datetime.utcfromtimestamp(time)
        return tstamp.strftime("%Y-%m-%dT%H:%M:%S") + ".%03d" % (tstamp.microsecond / 1000) + "Z"

    @classmethod
    def format_exception(cls, exc_info):
        return ''.join(traceback.format_exception(*exc_info)) if exc_info else ''

    @classmethod
    def serialize(cls, message):
        if sys.version_info < (3, 0):
            return json.dumps(message)
        else:
            return bytes(json.dumps(message, default=str), 'utf-8')

class LogstashFormatterVersion0(LogstashFormatterBase):
    version = 0

    def format(self, record):
        # Create message dict
        message = {
            '@timestamp': self.format_timestamp(record.created),
            '@message': record.getMessage(),
            '@source': self.format_source(self.message_type, self.host,
                                          record.pathname),
            '@source_host': self.host,
            '@source_path': record.pathname,
            '@tags': self.tags,
            '@type': self.message_type,
            '@fields': {
                'levelname': record.levelname,
                'logger': record.name,
            },
        }

        # Add extra fields
        # message['@fields'].update(self.get_extra_fields(record))

        # If exception, add debug info
        if record.exc_info:
            message['@fields'].update(self.get_debug_fields(record))

        return self.serialize(message)


class LogstashFormatterVersion1(LogstashFormatterBase):

    def format(self, record):
        # Create message dict
        message = {
            '@timestamp': self.format_timestamp(record.created),
            '@version': '1',
            'message': record.getMessage(),
            'host_name': self.host,
            'path': record.pathname,
            'tags': self.tags,
            'type': self.message_type,

            # Extra Fields
            'level': record.levelname,
            'logger_name': record.name,
            'lineno': record.lineno,
            'process_no': record.process,
            'thread_name': record.threadName,
            'filename': record.filename,
            'module': record.module,
            'funcName': record.funcName,
        }

        # Add extra fields
        # message.update(self.get_extra_fields(record))

        # If exception, add debug info
        if record.exc_info:
            message.update(self.get_debug_fields(record))

        return self.serialize(message)



from logging.handlers import DatagramHandler, SocketHandler
from pysailfish.internal.observability.logstash import formatter


# Derive from object to force a new-style class and thus allow super() to work
# on Python 2.6
class TCPLogstashHandler(SocketHandler, object):
    """Python logging handler for Logstash. Sends events over TCP.
    :param host: The host of the logstash server.
    :param port: The port of the logstash server (default 5959).
    :param message_type: The type of the message (default logstash).
    :param fqdn; Indicates whether to show fully qualified domain name or not (default False).
    :param version: version of logstash event schema (default is 0).
    :param tags: list of tags for a logger (default is None).
    """

    def __init__(self, host, port=5959, message_type='logstash', tags=None, fqdn=False, version=0):
        super(TCPLogstashHandler, self).__init__(host, port)
        if version == 1:
            self.formatter = formatter.LogstashFormatterVersion1(message_type, tags, fqdn)
        else:
            self.formatter = formatter.LogstashFormatterVersion0(message_type, tags, fqdn)

    def makePickle(self, record):
        return self.formatter.format(record) + b'\n'



import logging
import threading
import signal

import pysailfish.internal.observability.log_helper as lh
import pysailfish.internal.devTools.AtomicData as AtomicData

class ThreadClass(object):
    def __init__(self):
        self.__logger: logging.Logger = None
        self._is_running: AtomicData.AtomicData = None
        self._th: threading = None

    # override
    def init_component(self
                       , logger: logging.Logger):
        self.__logger = logger
        self._is_running = AtomicData.AtomicData()
        self._is_running.save(user_data=True)

        self._th = threading.Thread(target=self._main)
        signal.signal(signal.SIGINT, self._cleanup)

    # override
    def _user_main(self):
        raise NotImplementedError

    def start_th(self) -> None:
        self._th.start()

    def stop_th(self) -> None:
        if not self._is_running.get():
            return

        self._is_running.save(user_data=False)
        self._th.join()

    def _main(self):
        self.__logger.info("Start")

        while self._is_running.get():
            if not self._is_running.get():
                break

            # TODO - add user processing logic
            self._user_main()

            if not self._is_running.get():
                break

        self.__logger.info("End")

    def _cleanup(self, signum, frame):
        self.stop_th()
        exit(0)



from logging.handlers import DatagramHandler, SocketHandler
from pysailfish.internal.observability.logstash.handler_tcp import TCPLogstashHandler
from pysailfish.internal.observability.logstash import formatter

class UDPLogstashHandler(TCPLogstashHandler, DatagramHandler):
    """Python logging handler for Logstash. Sends events over UDP.
    :param host: The host of the logstash server.
    :param port: The port of the logstash server (default 5959).
    :param message_type: The type of the message (default logstash).
    :param fqdn; Indicates whether to show fully qualified domain name or not (default False).
    :param version: version of logstash event schema (default is 0).
    :param tags: list of tags for a logger (default is None).
    """

    def makePickle(self, record):
        return self.formatter.format(record)


# For backward compatibility
LogstashHandler = UDPLogstashHandler




import logging
import threading
from pysailfish.internal.observability.logging_json import JSONFormatter
from pysailfish.internal.observability.logstash.handler_tcp import TCPLogstashHandler
from pysailfish.internal.observability.logstash.handler_udp import UDPLogstashHandler, LogstashHandler
from logging.handlers import TimedRotatingFileHandler

__logger_singleton_lock = threading.Lock()

def __init_logger_imp(logger_name: str = "default_logger"
                      , log_level: int = logging.DEBUG
                      , is_print_to_file: bool = False
                      , log_file_root: str = "./"
                      , is_print_to_console: bool = True
                      , is_json_output: bool = True
                      , is_print_to_logstash: bool = False
                      , logstash_host: str = None
                      , logstash_port: int = None
                      , logstash_user_tags: list = None) -> logging.Logger:
    logger = logging.getLogger(logger_name)
    logger.setLevel(log_level)

    formatter = logging.Formatter('%(asctime)s [%(levelname)s] [%(name)s] [%(filename)s:%(funcName)s:%(lineno)d] [%(threadName)s:%(process)d] %(message)s')
    if is_json_output:
        formatter = JSONFormatter(fields={
            "log_time": "asctime"
            , "level_name": "levelname"
            , "logger_name": "name"
            , "file_name": "filename"
            , "func_name": "funcName"
            , "line_no": "lineno"
            , "thread_name": "threadName"
            , "process_name": "process"
           })

    if is_print_to_file:
        fh = TimedRotatingFileHandler(f"{log_file_root}/{logger_name}.log"
                                      , when='h'
                                      , interval=1
                                      , backupCount=0
                                      , encoding=None
                                      , delay=False
                                      , utc=True
                                      )
        fh.setLevel(log_level)
        fh.setFormatter(formatter)
        logger.addHandler(fh)
    if is_print_to_console:
        ch = logging.StreamHandler()
        ch.setLevel(logging.DEBUG)
        ch.setFormatter(formatter)
        logger.addHandler(ch)
    if is_print_to_logstash:
        logger.addHandler(TCPLogstashHandler(logstash_host, logstash_port, tags=logstash_user_tags, version=1))

    return logger

def init_logger(logger_name: str = "default_logger"
                , log_level: int = logging.INFO
                , is_print_to_file: bool = False
                , log_file_root: str = "./"
                , is_print_to_console: bool = True
                , is_json_output: bool = True
                , is_print_to_logstash: bool = False
                , logstash_host: str = None
                , logstash_port: int = None
                , logstash_user_tags: list = None) -> logging.Logger:
    logger = logging.getLogger(logger_name)
    # double check lock for singleton
    if not logger.hasHandlers():
        with __logger_singleton_lock:
            if not logger.hasHandlers():
                return __init_logger_imp(logger_name=logger_name
                                         , log_level=log_level
                                         , is_print_to_file=is_print_to_file
                                         , log_file_root=log_file_root
                                         , is_print_to_console=is_print_to_console
                                         , is_json_output=is_json_output
                                         , is_print_to_logstash=is_print_to_logstash
                                         , logstash_host=logstash_host
                                         , logstash_port=logstash_port
                                         , logstash_user_tags=logstash_user_tags)
            else:
                return logger
    else:
        return logger



from typing import Dict, List, Set, Tuple, Union, Any
from datetime import datetime

import pysailfish.internal.MT_EA.mt4_const as mc
from pysailfish.internal.DataCtrl.MT4DataCtrl.MT4DataCtrl import MT4DataCtrl

class market_info(object):
    def __init__(self):
        self.__data_ctrl: MT4DataCtrl = None
        self.__fun_num_map: Dict[str, int] = dict()

    def init_component(self, data_ctrl: MT4DataCtrl, fun_num_map: Dict[str, int]):
        self.__data_ctrl = data_ctrl
        self.__fun_num_map = fun_num_map

    def market_info(self, symbol: str, type: int) -> float:
        self.__data_ctrl.add_data_to_msg(data=str(self.__fun_num_map["MarketInfo"]))
        self.__data_ctrl.add_data_to_msg(data=symbol)
        self.__data_ctrl.add_data_to_msg(data=str(type))
        self.__data_ctrl.send_message()

        return self.__data_ctrl.receive_float_from_mt4()

    def symbols_total(self, selected: bool) -> int:
        self.__data_ctrl.add_data_to_msg(data=str(self.__fun_num_map["SymbolsTotal"]))
        self.__data_ctrl.add_data_to_msg(data=str(selected))
        self.__data_ctrl.send_message()

        return self.__data_ctrl.receive_int_from_mt4()

    def symbol_name(self, pos: int, selected: bool) -> str:
        self.__data_ctrl.add_data_to_msg(data=str(self.__fun_num_map["SymbolName"]))
        self.__data_ctrl.add_data_to_msg(data=str(pos))
        self.__data_ctrl.add_data_to_msg(data=str(selected))
        self.__data_ctrl.send_message()

        return self.__data_ctrl.receive_string_from_mt4()

    def symbol_select(self, name: str, select: bool) -> bool:
        self.__data_ctrl.add_data_to_msg(data=str(self.__fun_num_map["SymbolSelect"]))
        self.__data_ctrl.add_data_to_msg(data=name)
        self.__data_ctrl.add_data_to_msg(data=str(select))
        self.__data_ctrl.send_message()

        return self.__data_ctrl.receive_bool_from_mt4()

    def symbol_info_double(self, name: str, prop_id: int) -> float:
        self.__data_ctrl.add_data_to_msg(data=str(self.__fun_num_map["SymbolInfoDouble"]))
        self.__data_ctrl.add_data_to_msg(data=name)
        self.__data_ctrl.add_data_to_msg(data=str(prop_id))
        self.__data_ctrl.send_message()

        return self.__data_ctrl.receive_float_from_mt4()

    def symbol_info_integer(self, name: str, prop_id: int) -> int:
        self.__data_ctrl.add_data_to_msg(data=str(self.__fun_num_map["SymbolInfoInteger"]))
        self.__data_ctrl.add_data_to_msg(data=name)
        self.__data_ctrl.add_data_to_msg(data=str(prop_id))
        self.__data_ctrl.send_message()

        return self.__data_ctrl.receive_int_from_mt4()

    def symbol_info_string(self, name: str, prop_id: int) -> str:
        self.__data_ctrl.add_data_to_msg(data=str(self.__fun_num_map["SymbolInfoString"]))
        self.__data_ctrl.add_data_to_msg(data=name)
        self.__data_ctrl.add_data_to_msg(data=str(prop_id))
        self.__data_ctrl.send_message()

        return self.__data_ctrl.receive_string_from_mt4()

    def symbol_info_tick(self, symbol: str) -> Union[datetime, float, float, float, int]:
        self.__data_ctrl.add_data_to_msg(data=str(self.__fun_num_map["SymbolInfoTick"]))
        self.__data_ctrl.add_data_to_msg(data=symbol)
        self.__data_ctrl.send_message()

        str_ret: str = self.__data_ctrl.receive_string_from_mt4()

        parts = str_ret.split(",")

        return (datetime.strptime(parts[0], "%Y.%m.%d %H:%M:%S")
                , float(parts[1])
                , float(parts[2])
                , float(parts[3])
                , int(parts[4]))

    def symbol_info_session_quote(self, name: str, day_of_week: int, session_index: int) -> Union[datetime, datetime]:
        self.__data_ctrl.add_data_to_msg(data=str(self.__fun_num_map["SymbolInfoSessionQuote"]))
        self.__data_ctrl.add_data_to_msg(data=name)
        self.__data_ctrl.add_data_to_msg(data=str(day_of_week))
        self.__data_ctrl.add_data_to_msg(data=str(session_index))
        self.__data_ctrl.send_message()

        str_ret: str = self.__data_ctrl.receive_string_from_mt4()

        parts = str_ret.split(",")

        return (datetime.strptime(parts[0], "%Y.%m.%d %H:%M:%S")
                , datetime.strptime(parts[1], "%Y.%m.%d %H:%M:%S"))

    def symbol_info_session_trade(self, name: str, day_of_week: int, session_index: int) -> Union[datetime, datetime]:
        self.__data_ctrl.add_data_to_msg(data=str(self.__fun_num_map["SymbolInfoSessionTrade"]))
        self.__data_ctrl.add_data_to_msg(data=name)
        self.__data_ctrl.add_data_to_msg(data=str(day_of_week))
        self.__data_ctrl.add_data_to_msg(data=str(session_index))
        self.__data_ctrl.send_message()

        str_ret: str = self.__data_ctrl.receive_string_from_mt4()

        parts = str_ret.split(",")

        return (datetime.strptime(parts[0], "%Y.%m.%d %H:%M:%S")
                , datetime.strptime(parts[1], "%Y.%m.%d %H:%M:%S"))



import collections
import datetime
import json
import logging
from typing import Any, Dict

standard_attributes = (
    "name",
    "msg",
    "args",
    "levelname",
    "levelno",
    "pathname",
    "filename",
    "module",
    "exc_info",
    "exc_text",
    "stack_info",
    "lineno",
    "funcName",
    "created",
    "msecs",
    "relativeCreated",
    "thread",
    "threadName",
    "processName",
    "process",
    "message",
    "asctime",
)


def _extra_attributes(record: logging.LogRecord) -> Dict[str, Any]:
    return {
        name: record.__dict__[name]
        for name in set(record.__dict__).difference(standard_attributes)
    }


def _value(record: logging.LogRecord, field_name_or_value: Any) -> Any:
    """
    Retrieve value from record if possible. Otherwise use value.
    :param record: The record to extract a field named as in field_name_or_value.
    :param field_name_or_value: The field name to extract from record or the default value to use if not present.
    """
    try:
        return getattr(record, field_name_or_value)
    except:
        return field_name_or_value


def default_converter(obj: Any) -> str:
    if isinstance(obj, datetime.datetime):
        return obj.isoformat()
    return str(obj)


class JSONFormatter(logging.Formatter):
    def __init__(
        self,
        *args,
        fields: Dict[str, Any] = None,
        message_field_name: str = "msg",
        **kwargs,
    ):
        # Allow to provide any formatter setting (useful to provide a custom date format)
        super().__init__(*args, **kwargs)
        self.fields = fields or {}
        self.usesTime = lambda: "asctime" in self.fields.values()
        self.message_field_name = message_field_name

    def format(self, record: logging.LogRecord):
        # Let python set every additional record field
        super().format(record)

        message = {
            field_name: _value(record, field_value)
            for field_name, field_value in self.fields.items()
        }
        if isinstance(record.msg, collections.abc.Mapping):
            message.update(record.msg)
        else:
            message[self.message_field_name] = super().formatMessage(record)

        message.update(_extra_attributes(record))

        if record.exc_info:
            message["exception"] = {
                "type": record.exc_info[0].__name__,
                "message": str(record.exc_info[1]),
                "stack": self.formatException(record.exc_info),
            }

        if len(message) == 1 and self.message_field_name in message:
            return super().formatMessage(record)

        return json.dumps(message, default=default_converter)

    def formatMessage(self, record: logging.LogRecord) -> str:
        # Speed up this step by doing nothing
        return ""




# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['pysailfish',
 'pysailfish.app',
 'pysailfish.app.EAs',
 'pysailfish.internal',
 'pysailfish.internal.DataCtrl.MT4DataCtrl',
 'pysailfish.internal.MT_EA',
 'pysailfish.internal.Network',
 'pysailfish.internal.devTools',
 'pysailfish.internal.observability',
 'pysailfish.internal.observability.logging_json',
 'pysailfish.internal.observability.logstash']

package_data = \
{'': ['*']}

setup_kwargs = {
    'name': 'pysailfish',
    'version': '1.9.0',
    'description': '',
    'long_description': '\n## pysailfish\n\n### Example EA --- simple client\n```python\nfrom pysailfish.internal.MT_EA.MT4_EA import MT4_EA\n\nclass MT4DemoEA(MT4_EA):\n    def __init__(self):\n        super().__init__()\n\n    # override\n    def _OnInit(self) -> int:\n        self._logger.info(self._user_inputs)\n        self._logger.info("Here")\n        return 0\n\n    # override\n    def _OnDeinit(self, reason: int) -> None:\n        self._logger.info(f"Here reason: {reason}")\n        return None\n\n    # override\n    def _OnTick(self) -> None:\n        self._logger.info("Here")\n        return None\n\n    # override\n    def _OnTimer(self) -> None:\n        return None\n\n    # override\n    def _OnTester(self) -> float:\n        self._logger.info("Here")\n        return 0.0\n\n    # override\n    def _OnChartEvent(self\n                      , id: int\n                      , lparam: int\n                      , dparam: float\n                      , sparam: str) -> None:\n        self._logger.info(f"Here id: {id} lparam: {lparam} dparam: {dparam} sparam: {sparam}")\n        return None\n\ndef main() -> None:\n    ea = MT4DemoEA()\n    ea.InitComponent(server_ip="127.0.0.1"\n                     , server_port=23456\n                     , ea_name="MT4DemoEA")\n    ea.StartEA()\n\nif __name__ == "__main__":\n    main()\n```\n\n### Example EA --- Moving Average EA\n```python\nfrom datetime import datetime\n\nimport pysailfish.internal.MT_EA.mt4_const as mc\nfrom pysailfish.internal.MT_EA.MT4_EA import MT4_EA\n\nclass MA_EA(MT4_EA):\n    def __init__(self):\n        super().__init__()\n\n    # override\n    def _OnInit(self) -> int:\n        self._logger.info(self._user_inputs)\n        self._logger.info("Here")\n        return 0\n\n    # override\n    def _OnDeinit(self, reason: int) -> None:\n        self._logger.info(f"Here reason: {reason}")\n        return None\n\n    # override\n    def _OnTick(self) -> None:\n        vv = self.iADX(symbol=self._pv.symbol\n                       , timeframe=mc.PERIOD_CURRENT\n                       , period=self._user_inputs["MovingPeriod"]\n                       , applied_price=mc.PRICE_CLOSE\n                       , mode=mc.MODE_MAIN\n                       , shift=0)\n        # self._logger.info(f"Here {self._pv.symbol} {self._pv.time[0]}")\n        # --- check for history and trading\n        if self._pv.bars < 100 or self._pv.is_trade_allowed == False:\n            return\n\n        # --- calculate open orders by current symbol\n        if self.__calculate_current_orders(self._pv.symbol) == 0:\n            self.__check_for_open()\n        else:\n            self.__check_for_close()\n        return None\n\n    # override\n    def _OnTimer(self) -> None:\n        return None\n\n    # override\n    def _OnTester(self) -> float:\n        self._logger.info("Here")\n        return 0.0\n\n    # override\n    def _OnChartEvent(self\n                      , id: int\n                      , lparam: int\n                      , dparam: float\n                      , sparam: str) -> None:\n        self._logger.info(f"Here id: {id} lparam: {lparam} dparam: {dparam} sparam: {sparam}")\n        return None\n\n    def __check_for_close(self) -> None:\n        ma: float = 0\n        #--- go trading only for first tiks of new bar\n        if self._pv.volume[0] > 1:\n            return None\n        #--- get Moving Average\n        ma = self.iMA(symbol=self._pv.symbol\n                      , timeframe=mc.PERIOD_CURRENT\n                      , ma_period=self._user_inputs["MovingPeriod"]\n                      , ma_shift=self._user_inputs["MovingShift"]\n                      , ma_method=mc.MODE_SMA\n                      , applied_price=mc.PRICE_CLOSE\n                      , shift=0)\n        #---\n        for i in range(self.OrdersTotal()):\n            if self.OrderSelect(index=i, select=mc.SELECT_BY_POS, pool=mc.MODE_TRADES) == False:\n                break\n            if self.OrderMagicNumber() != self._pv.magic_num or self.OrderSymbol() != self._pv.symbol:\n                continue\n            #--- check order type\n            if self.OrderType() == mc.OP_BUY:\n                if self._pv.open[1] > ma and self._pv.close[1] < ma:\n                    if not self.OrderClose(ticket=self.OrderTicket(), lots=self.OrderLots(), price=self._pv.bid, slippage=3, arrow_color=mc.clrWhite):\n                        self._logger.error(f"OrderClose error {self.GetLastError()}")\n                break\n            if self.OrderType() == mc.OP_SELL:\n                if self._pv.open[1] < ma and self._pv.close[1] > ma:\n                    if not self.OrderClose(ticket=self.OrderTicket(), lots=self.OrderLots(), price=self._pv.ask, slippage=3, arrow_color=mc.clrWhite):\n                        self._logger.error(f"OrderClose error {self.GetLastError()}");\n                break\n\n    def __lots_optimized(self) -> float:\n        lot: float = float(self._user_inputs["Lots"])\n        maximum_risk: float = float(self._user_inputs["MaximumRisk"])\n        decrease_factor: float = float(self._user_inputs["DecreaseFactor"])\n        orders: int = self.OrdersHistoryTotal() # history orders total\n        losses: int = 0 # number of losses orders without a break\n        #--- select lot size\n        lot = self.NormalizeDouble(value=(self.AccountFreeMargin() * maximum_risk / 1000.0), digits=1);\n        #--- calcuulate number of losses orders without a break\n        if decrease_factor > 0:\n            for i in reversed(range(orders)):\n                if self.OrderSelect(index=i, select=mc.SELECT_BY_POS, pool=mc.MODE_HISTORY) == False:\n                    self._logger.error("Error in history")\n                    break\n                if self.OrderSymbol() != self._pv.symbol or self.OrderType() > mc.OP_SELL:\n                    continue\n                #---\n                if self.OrderProfit() > 0:\n                    break\n                if self.OrderProfit() < 0:\n                    losses += 1\n            if losses > 1:\n                lot = self.NormalizeDouble(value=(lot - lot * losses / decrease_factor), digits=1)\n        #--- return lot size\n        if lot < 0.1:\n            lot = 0.1\n        return lot\n\n    def __check_for_open(self) -> None:\n        ma: float = 0\n        res: int = 0\n        #--- go trading only for first tiks of new bar\n        if self._pv.volume[0] > 1:\n            return None\n        #--- get Moving Average\n        ma = self.iMA(symbol=self._pv.symbol\n                      , timeframe=mc.PERIOD_CURRENT\n                      , ma_period=self._user_inputs["MovingPeriod"]\n                      , ma_shift=self._user_inputs["MovingShift"]\n                      , ma_method=mc.MODE_SMA\n                      , applied_price=mc.PRICE_CLOSE\n                      , shift=0)\n        #--- sell conditions\n        if self._pv.open[1] > ma and self._pv.close[1] < ma:\n            res = self.OrderSend(symbol=self._pv.symbol\n                                 , cmd=mc.OP_SELL\n                                 , volume=self.__lots_optimized()\n                                 , price=self._pv.bid\n                                 , slippage=3\n                                 , stoploss=0\n                                 , takeprofit=0\n                                 , comment=""\n                                 , magic=self._pv.magic_num\n                                 , expiration=datetime(1970, 1, 1, 0, 0, 0)\n                                 , arrow_color=mc.clrRed)\n            return None\n        #--- buy conditions\n        if self._pv.open[1] < ma and self._pv.close[1] > ma:\n            res = self.OrderSend(symbol=self._pv.symbol\n                                 , cmd=mc.OP_BUY\n                                 , volume=self.__lots_optimized()\n                                 , price=self._pv.ask\n                                 , slippage=3\n                                 , stoploss=0\n                                 , takeprofit=0\n                                 , comment=""\n                                 , magic=self._pv.magic_num\n                                 , expiration=datetime(1970, 1, 1, 0, 0, 0)\n                                 , arrow_color=mc.clrBlue)\n            return None\n\n    def __calculate_current_orders(self, symbol: str) -> int:\n        buys: int = 0\n        sells: int = 0\n        # ---\n        for i in range(self.OrdersTotal()):\n            if self._tf.order_select(index=i, select=mc.SELECT_BY_POS, pool=mc.MODE_TRADES) == False:\n                break\n            if self.OrderSymbol() == self._pv.symbol and self.OrderMagicNumber() == self._pv.magic_num:\n                if self.OrderType() == mc.OP_BUY:\n                    buys += 1\n                if self.OrderType() == mc.OP_SELL:\n                    sells += 1\n        # --- return orders volume\n        if buys > 0:\n            return buys\n        else:\n            return -sells\n\ndef main() -> None:\n    ea = MA_EA()\n    ea.InitComponent(server_ip="127.0.0.1"\n                     , server_port=23456\n                     , ea_name="MA_EA")\n    ea.StartEA()\n\nif __name__ == "__main__":\n    main()\n```\n\n### How to publish to pypi\n```bash\n# set up pypi token\n$ poetry config pypi-token.pypi my-token\n\n# build the project\n$ poetry build\n\n# publish the project\n$ poetry publish\n\n# DONE\n```\n',
    'author': 'SulfredLee',
    'author_email': 'sflee1112@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



from typing import Dict, List, Set, Tuple, Union, Any
from datetime import datetime

import pysailfish.internal.MT_EA.mt4_const as mc
from pysailfish.internal.DataCtrl.MT4DataCtrl.MT4DataCtrl import MT4DataCtrl

# https://docs.mql4.com/objects/objectname
class object_functions(object):
    def __init__(self):
        self.__data_ctrl: MT4DataCtrl = None
        self.__fun_num_map: Dict[str, int] = dict()

    def init_component(self, data_ctrl: MT4DataCtrl, fun_num_map: Dict[str, int]):
        self.__data_ctrl = data_ctrl
        self.__fun_num_map = fun_num_map

    # chart_id: int, object_name: str, object_type: int, sub_window: int, time1: datetime, price1: float ... timeN: datetime, priceN: float
    def object_create_001(self, **kwargs) -> bool:
        self.__data_ctrl.add_data_to_msg(data=str(self.__fun_num_map["ObjectCreate_001"]))
        for k, v in kwargs.items():
            if isinstance(v, datetime):
                self.__data_ctrl.add_data_to_msg(data=v.strftime("%Y.%m.%d %H:%M:%S"))
            else:
                self.__data_ctrl.add_data_to_msg(data=v)
        self.__data_ctrl.send_message()

        return self.__data_ctrl.receive_bool_from_mt4()

    def object_create_002(self
                          , object_name: str
                          , object_type: int
                          , sub_window: int
                          , time1: datetime
                          , price1: float
                          , time2: datetime = datetime(1970, 1, 1, 0, 0, 0)
                          , price2: float = 0
                          , time3: datetime = datetime(1970, 1, 1, 0, 0, 0)
                          , price3: float = 0) -> bool:
        self.__data_ctrl.add_data_to_msg(data=str(self.__fun_num_map["ObjectCreate_002"]))
        self.__data_ctrl.add_data_to_msg(data=object_name)
        self.__data_ctrl.add_data_to_msg(data=object_type)
        self.__data_ctrl.add_data_to_msg(data=str(sub_window))
        self.__data_ctrl.add_data_to_msg(data=time1.strftime("%Y.%m.%d %H:%M:%S"))
        self.__data_ctrl.add_data_to_msg(data=str(price1))
        self.__data_ctrl.add_data_to_msg(data=time2.strftime("%Y.%m.%d %H:%M:%S"))
        self.__data_ctrl.add_data_to_msg(data=str(price2))
        self.__data_ctrl.add_data_to_msg(data=time3.strftime("%Y.%m.%d %H:%M:%S"))
        self.__data_ctrl.add_data_to_msg(data=str(price3))
        self.__data_ctrl.send_message()

        return self.__data_ctrl.receive_bool_from_mt4()

    def object_name(self, object_index: int) -> str:
        self.__data_ctrl.add_data_to_msg(data=str(self.__fun_num_map["ObjectName"]))
        self.__data_ctrl.add_data_to_msg(data=str(object_index))
        self.__data_ctrl.send_message()

        return self.__data_ctrl.receive_string_from_mt4()

    def object_delete_001(self, chart_id: int, object_name: str) -> bool:
        self.__data_ctrl.add_data_to_msg(data=str(self.__fun_num_map["ObjectDelete_001"]))
        self.__data_ctrl.add_data_to_msg(data=str(chart_id))
        self.__data_ctrl.add_data_to_msg(data=object_name)
        self.__data_ctrl.send_message()

        return self.__data_ctrl.receive_bool_from_mt4()

    def object_delete_002(self, object_name: str) -> bool:
        self.__data_ctrl.add_data_to_msg(data=str(self.__fun_num_map["ObjectDelete_002"]))
        self.__data_ctrl.add_data_to_msg(data=object_name)
        self.__data_ctrl.send_message()

        return self.__data_ctrl.receive_bool_from_mt4()

    def objects_delete_all_001(self, chart_id: int, sub_window: int = -1, object_type: int = -1) -> int:
        self.__data_ctrl.add_data_to_msg(data=str(self.__fun_num_map["ObjectsDeleteAll_001"]))
        self.__data_ctrl.add_data_to_msg(data=str(chart_id))
        self.__data_ctrl.add_data_to_msg(data=str(sub_window))
        self.__data_ctrl.add_data_to_msg(data=str(object_type))
        self.__data_ctrl.send_message()

        return self.__data_ctrl.receive_int_from_mt4()

    def objects_delete_all_002(self, sub_window: int = -1, object_type: int = -1) -> int:
        self.__data_ctrl.add_data_to_msg(data=str(self.__fun_num_map["ObjectsDeleteAll_002"]))
        self.__data_ctrl.add_data_to_msg(data=str(sub_window))
        self.__data_ctrl.add_data_to_msg(data=str(object_type))
        self.__data_ctrl.send_message()

        return self.__data_ctrl.receive_int_from_mt4()

    def objects_delete_all_003(self, chart_id: int, prefix: str, sub_window: int = -1, object_type: int = -1) -> int:
        self.__data_ctrl.add_data_to_msg(data=str(self.__fun_num_map["ObjectsDeleteAll_003"]))
        self.__data_ctrl.add_data_to_msg(data=str(chart_id))
        self.__data_ctrl.add_data_to_msg(data=prefix)
        self.__data_ctrl.add_data_to_msg(data=str(sub_window))
        self.__data_ctrl.add_data_to_msg(data=str(object_type))
        self.__data_ctrl.send_message()

        return self.__data_ctrl.receive_int_from_mt4()

    def object_find_001(self, chart_id: int, object_name: str) -> int:
        self.__data_ctrl.add_data_to_msg(data=str(self.__fun_num_map["ObjectFind_001"]))
        self.__data_ctrl.add_data_to_msg(data=str(chart_id))
        self.__data_ctrl.add_data_to_msg(data=object_name)
        self.__data_ctrl.send_message()

        return self.__data_ctrl.receive_int_from_mt4()

    def object_find_002(self, object_name: str) -> int:
        self.__data_ctrl.add_data_to_msg(data=str(self.__fun_num_map["ObjectFind_002"]))
        self.__data_ctrl.add_data_to_msg(data=object_name)
        self.__data_ctrl.send_message()

        return self.__data_ctrl.receive_int_from_mt4()

    def object_get_time_by_value(self, chart_id: int, object_name: str, value: float, line_id: int = 0) -> datetime:
        self.__data_ctrl.add_data_to_msg(data=str(self.__fun_num_map["ObjectGetTimeByValue"]))
        self.__data_ctrl.add_data_to_msg(data=str(chart_id))
        self.__data_ctrl.add_data_to_msg(data=object_name)
        self.__data_ctrl.add_data_to_msg(data=str(value))
        self.__data_ctrl.add_data_to_msg(data=str(line_id))
        self.__data_ctrl.send_message()

        return self.__data_ctrl.receive_datetime_from_mt4()

    def object_get_value_by_time(self, chart_id: int, object_name: str, time: datetime, line_id: int = 0) -> float:
        self.__data_ctrl.add_data_to_msg(data=str(self.__fun_num_map["ObjectGetValueByTime"]))
        self.__data_ctrl.add_data_to_msg(data=str(chart_id))
        self.__data_ctrl.add_data_to_msg(data=object_name)
        self.__data_ctrl.add_data_to_msg(data=time.strftime("%Y.%m.%d %H:%M:%S"))
        self.__data_ctrl.add_data_to_msg(data=str(line_id))
        self.__data_ctrl.send_message()

        return self.__data_ctrl.receive_float_from_mt4()

    def object_move(self, object_name: str, point_index: int, time: datetime, price: float) -> bool:
        self.__data_ctrl.add_data_to_msg(data=str(self.__fun_num_map["ObjectMove"]))
        self.__data_ctrl.add_data_to_msg(data=object_name)
        self.__data_ctrl.add_data_to_msg(data=str(point_index))
        self.__data_ctrl.add_data_to_msg(data=time.strftime("%Y.%m.%d %H:%M:%S"))
        self.__data_ctrl.add_data_to_msg(data=str(price))
        self.__data_ctrl.send_message()

        return self.__data_ctrl.receive_bool_from_mt4()

    def objects_total_001(self, chart_id: int, sub_window: int = -1, type: int = -1) -> int:
        self.__data_ctrl.add_data_to_msg(data=str(self.__fun_num_map["ObjectsTotal_001"]))
        self.__data_ctrl.add_data_to_msg(data=str(chart_id))
        self.__data_ctrl.add_data_to_msg(data=str(sub_windows))
        self.__data_ctrl.add_data_to_msg(data=str(type))
        self.__data_ctrl.send_message()

        return self.__data_ctrl.receive_int_from_mt4()

    def objects_total_002(self, type: int = -1) -> int:
        self.__data_ctrl.add_data_to_msg(data=str(self.__fun_num_map["ObjectsTotal_002"]))
        self.__data_ctrl.add_data_to_msg(data=str(type))
        self.__data_ctrl.send_message()

        return self.__data_ctrl.receive_int_from_mt4()

    def object_get_double(self, chart_id: int, object_name: str, prop_id: int, prop_modifier: int = 0) -> float:
        self.__data_ctrl.add_data_to_msg(data=str(self.__fun_num_map["ObjectGetDouble"]))
        self.__data_ctrl.add_data_to_msg(data=str(chart_id))
        self.__data_ctrl.add_data_to_msg(data=object_name)
        self.__data_ctrl.add_data_to_msg(data=str(prop_id))
        self.__data_ctrl.add_data_to_msg(data=str(prop_modifier))
        self.__data_ctrl.send_message()

        return self.__data_ctrl.receive_float_from_mt4()

    def object_get_integer(self, chart_id: int, object_name: str, prop_id: int, prop_modifier: int = 0) -> int:
        self.__data_ctrl.add_data_to_msg(data=str(self.__fun_num_map["ObjectGetInteger"]))
        self.__data_ctrl.add_data_to_msg(data=str(chart_id))
        self.__data_ctrl.add_data_to_msg(data=object_name)
        self.__data_ctrl.add_data_to_msg(data=str(prop_id))
        self.__data_ctrl.add_data_to_msg(data=str(prop_modifier))
        self.__data_ctrl.send_message()

        return self.__data_ctrl.receive_int_from_mt4()

    def object_get_string(self, chart_id: int, object_name: str, prop_id: int, prop_modifier: int = 0) -> str:
        self.__data_ctrl.add_data_to_msg(data=str(self.__fun_num_map["ObjectGetString"]))
        self.__data_ctrl.add_data_to_msg(data=str(chart_id))
        self.__data_ctrl.add_data_to_msg(data=object_name)
        self.__data_ctrl.add_data_to_msg(data=str(prop_id))
        self.__data_ctrl.add_data_to_msg(data=str(prop_modifier))
        self.__data_ctrl.send_message()

        return self.__data_ctrl.receive_string_from_mt4()

    def object_set_double_001(self, chart_id: int, object_name: str, prop_id: int, prop_value: float) -> bool:
        self.__data_ctrl.add_data_to_msg(data=str(self.__fun_num_map["ObjectSetDouble_001"]))
        self.__data_ctrl.add_data_to_msg(data=str(chart_id))
        self.__data_ctrl.add_data_to_msg(data=object_name)
        self.__data_ctrl.add_data_to_msg(data=str(prop_id))
        self.__data_ctrl.add_data_to_msg(data=str(prop_value))
        self.__data_ctrl.send_message()

        return self.__data_ctrl.receive_bool_from_mt4()

    def object_set_double_002(self, chart_id: int, object_name: str, prop_id: int, prop_modifier: int, prop_value: float) -> bool:
        self.__data_ctrl.add_data_to_msg(data=str(self.__fun_num_map["ObjectSetDouble_002"]))
        self.__data_ctrl.add_data_to_msg(data=str(chart_id))
        self.__data_ctrl.add_data_to_msg(data=object_name)
        self.__data_ctrl.add_data_to_msg(data=str(prop_id))
        self.__data_ctrl.add_data_to_msg(data=str(prop_modifier))
        self.__data_ctrl.add_data_to_msg(data=str(prop_value))
        self.__data_ctrl.send_message()

        return self.__data_ctrl.receive_bool_from_mt4()

    def object_set_integer_001(self, chart_id: int, object_name: str, prop_id: int, prop_value: float) -> bool:
        self.__data_ctrl.add_data_to_msg(data=str(self.__fun_num_map["ObjectSetInteger_001"]))
        self.__data_ctrl.add_data_to_msg(data=str(chart_id))
        self.__data_ctrl.add_data_to_msg(data=object_name)
        self.__data_ctrl.add_data_to_msg(data=str(prop_id))
        self.__data_ctrl.add_data_to_msg(data=str(prop_value))
        self.__data_ctrl.send_message()

        return self.__data_ctrl.receive_bool_from_mt4()

    def object_set_integer_002(self, chart_id: int, object_name: str, prop_id: int, prop_modifier: int, prop_value: float) -> bool:
        self.__data_ctrl.add_data_to_msg(data=str(self.__fun_num_map["ObjectSetInteger_002"]))
        self.__data_ctrl.add_data_to_msg(data=str(chart_id))
        self.__data_ctrl.add_data_to_msg(data=object_name)
        self.__data_ctrl.add_data_to_msg(data=str(prop_id))
        self.__data_ctrl.add_data_to_msg(data=str(prop_modifier))
        self.__data_ctrl.add_data_to_msg(data=str(prop_value))
        self.__data_ctrl.send_message()

        return self.__data_ctrl.receive_bool_from_mt4()

    def object_set_string_001(self, chart_id: int, object_name: str, prop_id: int, prop_value: float) -> bool:
        self.__data_ctrl.add_data_to_msg(data=str(self.__fun_num_map["ObjectSetString_001"]))
        self.__data_ctrl.add_data_to_msg(data=str(chart_id))
        self.__data_ctrl.add_data_to_msg(data=object_name)
        self.__data_ctrl.add_data_to_msg(data=str(prop_id))
        self.__data_ctrl.add_data_to_msg(data=str(prop_value))
        self.__data_ctrl.send_message()

        return self.__data_ctrl.receive_bool_from_mt4()

    def object_set_string_002(self, chart_id: int, object_name: str, prop_id: int, prop_modifier: int, prop_value: float) -> bool:
        self.__data_ctrl.add_data_to_msg(data=str(self.__fun_num_map["ObjectSetString_002"]))
        self.__data_ctrl.add_data_to_msg(data=str(chart_id))
        self.__data_ctrl.add_data_to_msg(data=object_name)
        self.__data_ctrl.add_data_to_msg(data=str(prop_id))
        self.__data_ctrl.add_data_to_msg(data=str(prop_modifier))
        self.__data_ctrl.add_data_to_msg(data=str(prop_value))
        self.__data_ctrl.send_message()

        return self.__data_ctrl.receive_bool_from_mt4()

    def text_set_font(self, name: str, size: int, flags: int = 0, orientation: int = 0) -> bool:
        self.__data_ctrl.add_data_to_msg(data=str(self.__fun_num_map["TextSetFont"]))
        self.__data_ctrl.add_data_to_msg(data=name)
        self.__data_ctrl.add_data_to_msg(data=str(size))
        self.__data_ctrl.add_data_to_msg(data=str(flags))
        self.__data_ctrl.add_data_to_msg(data=str(orientation))
        self.__data_ctrl.send_message()

        return self.__data_ctrl.receive_bool_from_mt4()

    # def text_out(self, text: str, x: int, y: int, anchor: int, ) -> bool:
    #     self.__data_ctrl.add_data_to_msg(data=str(self.__fun_num_map["TextSetFont"]))
    #     self.__data_ctrl.add_data_to_msg(data=name)
    #     self.__data_ctrl.add_data_to_msg(data=str(size))
    #     self.__data_ctrl.add_data_to_msg(data=str(flags))
    #     self.__data_ctrl.add_data_to_msg(data=str(orientation))
    #     self.__data_ctrl.send_message()

    #     return self.__data_ctrl.receive_bool_from_mt4()

    def text_get_size(self, text: str) -> Union[int, int]:
        self.__data_ctrl.add_data_to_msg(data=str(self.__fun_num_map["TextGetSize"]))
        self.__data_ctrl.add_data_to_msg(data=text)
        self.__data_ctrl.send_message()

        str_ret: str = self.__data_ctrl.receive_string_from_mt4()
        parts = str_ret.split("-")

        return (int(parts[0]), int(parts[1])) # width, height

    def object_description(self, object_name: str) -> str:
        self.__data_ctrl.add_data_to_msg(data=str(self.__fun_num_map["ObjectDescription"]))
        self.__data_ctrl.add_data_to_msg(data=object_name)
        self.__data_ctrl.send_message()

        return self.__data_ctrl.receive_string_from_mt4()

    def object_get(self, object_name: str, index: int) -> float:
        self.__data_ctrl.add_data_to_msg(data=str(self.__fun_num_map["ObjectGet"]))
        self.__data_ctrl.add_data_to_msg(data=object_name)
        self.__data_ctrl.add_data_to_msg(data=str(index))
        self.__data_ctrl.send_message()

        return self.__data_ctrl.receive_float_from_mt4()

    def object_get_fibo_description(self, object_name: str, index: int) -> str:
        self.__data_ctrl.add_data_to_msg(data=str(self.__fun_num_map["ObjectGetFiboDescription"]))
        self.__data_ctrl.add_data_to_msg(data=object_name)
        self.__data_ctrl.add_data_to_msg(data=str(index))
        self.__data_ctrl.send_message()

        return self.__data_ctrl.receive_string_from_mt4()

    def object_get_shift_by_value(self, object_name: str, value: float) -> int:
        self.__data_ctrl.add_data_to_msg(data=str(self.__fun_num_map["ObjectGetShiftByValue"]))
        self.__data_ctrl.add_data_to_msg(data=object_name)
        self.__data_ctrl.add_data_to_msg(data=str(value))
        self.__data_ctrl.send_message()

        return self.__data_ctrl.receive_int_from_mt4()

    def object_get_value_by_shift(self, object_name: str, shift: int) -> float:
        self.__data_ctrl.add_data_to_msg(data=str(self.__fun_num_map["ObjectGetValueByShift"]))
        self.__data_ctrl.add_data_to_msg(data=object_name)
        self.__data_ctrl.add_data_to_msg(data=str(shift))
        self.__data_ctrl.send_message()

        return self.__data_ctrl.receive_float_from_mt4()

    def object_set(self, object_name: str, index: int, vlaue: float) -> bool:
        self.__data_ctrl.add_data_to_msg(data=str(self.__fun_num_map["ObjectSet"]))
        self.__data_ctrl.add_data_to_msg(data=object_name)
        self.__data_ctrl.add_data_to_msg(data=str(index))
        self.__data_ctrl.add_data_to_msg(data=str(value))
        self.__data_ctrl.send_message()

        return self.__data_ctrl.receive_bool_from_mt4()

    def object_set_fibo_description(self, object_name: str, index: int, text: str) -> bool:
        self.__data_ctrl.add_data_to_msg(data=str(self.__fun_num_map["ObjectSetFiboDescription"]))
        self.__data_ctrl.add_data_to_msg(data=object_name)
        self.__data_ctrl.add_data_to_msg(data=str(index))
        self.__data_ctrl.add_data_to_msg(data=text)
        self.__data_ctrl.send_message()

        return self.__data_ctrl.receive_bool_from_mt4()

    def object_set_text(self, object_name: str, text: str, font_size: int = 0, font_name: str = "", text_color: str = mc.clrBlack) -> bool:
        self.__data_ctrl.add_data_to_msg(data=str(self.__fun_num_map["ObjectSetText"]))
        self.__data_ctrl.add_data_to_msg(data=object_name)
        self.__data_ctrl.add_data_to_msg(data=text)
        self.__data_ctrl.add_data_to_msg(data=str(font_size))
        self.__data_ctrl.add_data_to_msg(data=font_name)
        self.__data_ctrl.add_data_to_msg(data=text_color)
        self.__data_ctrl.send_message()

        return self.__data_ctrl.receive_bool_from_mt4()

    def object_type(self, object_name: str) -> int:
        self.__data_ctrl.add_data_to_msg(data=str(self.__fun_num_map["ObjectType"]))
        self.__data_ctrl.add_data_to_msg(data=object_name)
        self.__data_ctrl.send_message()

        return self.__data_ctrl.receive_int_from_mt4()



import threading
from typing import List, Dict, Tuple, Any, Deque
import copy

class AtomicData(object):
    def __init__(self):
        self.__mutex: threading.Lock = threading.Lock()
        self.__user_data: Any = None

    def save(self, user_data: Any) -> None:
        with self.__mutex:
            self.__user_data = user_data

    def get(self) -> Any:
        with self.__mutex:
            return copy.deepcopy(self.__user_data)

    def is_same(self, other_data: Any) -> None:
        with self.__mutex:
            return self.__user_data == other_data
